#include "complex.h"
#include <stdio.h>
#include <string.h>

/////////////////////////////////////////////////////////////
/*
If the developer does not need to handle maximum values but
just display an error message just set the first parameter
in the routine 'setDivByZeroHandling' to 'false'.
*/
/////////////////////////////////////////////////////////////

//////////////// Error catching parameters //////////////////
#define HANDLE_DIVBYZERO    1
/////////////////////////////
double MAX_VAL_REAL = pow( 2, sizeof(double) * 8 ) - 1 ;
double MAX_VAL_IMAG = pow( 2, sizeof(double) * 8 ) - 1 ;

double err_div_subst_real = MAX_VAL_REAL, err_div_subst_imag = MAX_VAL_IMAG ;
int	errorDetectLevel = HANDLE_DIVBYZERO ;
/////////////////////////////
	class ParsingError
	{
	public:
		char message[256];
	};
///////////////////////////////////////////////////
///////////////////////////////////////////////////

//////////////// Error catching routines //////////////////
void setErrorDetectLevel( int errNo )
{
	errorDetectLevel |= errNo ;
}
	
void setDivByZeroHandling( bool ignore, double max_real, double max_imag )
{
	if( ignore )
	{        // non-standard behavior, ignore the error
        errorDetectLevel |= HANDLE_DIVBYZERO;
        err_div_subst_real = max_real ;
        err_div_subst_imag = max_imag ;
    } else            // throw an exception
        errorDetectLevel &= (!HANDLE_DIVBYZERO);
}

//////////////// constructors & destructors //////////////////////

complex::complex()
{
	real = imag = 0.0 ;
}

complex::complex( double r , double i )
{
	real = r ;		imag = i ;
}

complex::complex( double r )
{
	real = r ;	imag = 0.0 ;
}

complex::complex( const complex& c )
{
	real = c.real ;		imag = c.imag ;
}

complex::~complex()
{
}

// --------------- "Self" arithmethic -----------------------

void complex::operator+=( complex c_add )
{
	*this = *this + c_add ;
}

void complex::operator-=( complex c_sub )
{
	*this = *this - c_sub ;
}

void complex::operator*=( complex c_mul )
{
	*this = *this * c_mul ;
}

void complex::operator/=( complex c_div )
{
	*this = *this / c_div ;
}

// ----------  Addition -------------------------------------

complex operator+( complex c_add1 , complex c_add2 )
{
	return complex( c_add1.real + c_add2.real, c_add1.imag + c_add2.imag ) ;
}

complex operator+( complex c_add1 , double r )
{
	return c_add1 + complex( r , 0.0 );
}

complex operator+( double r , complex c_add2 )
{
	return complex( r , 0.0 ) + c_add2 ;
}

// ----------  Subtraction -------------------------------------

complex operator-( complex c )
{
	return (complex)( 0.0 - c ) ;
}

complex operator-( complex c_min1 , complex c_min2 )
{
	return complex( c_min1.real - c_min2.real, c_min1.imag - c_min2.imag ) ;
}

complex operator-( complex c_min1 , double r )
{
	return c_min1 - complex( r , 0.0 ) ;
}

complex operator-( double r , complex c_min2 )
{
	return complex( r , 0.0 ) - c_min2 ;
}

// ----------  Multiplication -------------------------------------

complex operator*( complex c_mul1 , complex c_mul2 )
{
	return complex ( c_mul1.real * c_mul2.real - c_mul1.imag * c_mul2.imag,
					 c_mul1.real * c_mul2.imag + c_mul1.imag * c_mul2.real ) ;
}

complex operator*( complex c_mul1 , double r )
{
	return c_mul1 * complex( r , 0.0 ) ;
}

complex operator*( double r , complex c_mul2 )
{
	return complex( r , 0.0 ) * c_mul2 ;
}

// ----------  Division -------------------------------------

complex operator/( complex c_dividend , complex c_divisor )
{
	complex q ;
	double r2 = c_divisor.real * c_divisor.real + c_divisor.imag * c_divisor.imag;

	if ( r2 == 0.0)
	{
		if( errorDetectLevel & HANDLE_DIVBYZERO )
		{
			q.real = err_div_subst_real ;
			q.imag = err_div_subst_imag ;
			return q;
		}
		else
		{
			ParsingError e ;
			strcpy(e.message, "Divide by zero");
			throw e;
		}
	}
	
	q.real = ( c_dividend.real * c_divisor.real + c_dividend.imag * c_divisor.imag ) / r2 ;
	q.imag = ( c_divisor.real * c_dividend.imag - c_dividend.real * c_divisor.imag ) / r2 ;
	
	return q ;
}

complex operator/(complex c_div1 , double c_divisor )
{
	return c_div1 / complex( c_divisor , 0.0 ); ;
}

complex operator/(double c_dividend , complex c_div2 )
{
	return complex( c_dividend , 0.0 ) / c_div2 ;
}

//////// Power //////////////////////////
complex pow( complex base , complex exponent )
{
	return exp( exponent * ln( base ) ) ;
}

complex pow( complex base , double exponent )
{
	return pow( base, complex( exponent, 0.0 ) ) ;
}

complex pow( complex base , int exponent )
{
	return pow( base, (double)exponent ) ;
}

complex sqrt( complex c )
{
	return pow( c, 0.5 ) ;
}

//////// Neperian Logarithm //////////////////////////
complex ln( complex c )
{
	double r = c.radius() ;

	return ( r == 0.0 ) ? complex( 0, 0 ) : complex( log( r ) , c.angle() ) ;
}

//////// Base 10 Logarithm //////////////////////////
complex log( complex c )
{	
	return ln( c ) / ln( complex( 10.0, 0.0 ) );
}

//////// Exponential ////////////////////////////////
complex exp( complex z )
{
	return complex( exp( z.real ) * cos( z.imag ), exp( z.real ) * sin( z.imag ) ) ;
}

//////// Trigonometric //////////////////////////////
complex sin( complex z )
{
	complex i( 0.0 , 1.0 ) ;
	
	return ( exp( i * z ) - exp( -i * z ) ) / ( 2.0 * i );
}

complex cos( complex z )
{
	complex i( 0.0 , 1.0 ) ;
	
	return ( exp( i * z ) + exp( -i * z ) ) / 2.0 ;
}

complex tg( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	complex e1_z = exp( i * z ) , e2_z = exp( -i * z ) ;
	
	return ( e1_z - e2_z ) / ( i * ( e1_z + e2_z ) ) ;
}

complex ctg( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	complex e1_z = exp( i * z ) , e2_z = exp( -i * z ) ;

	return ( e1_z + e2_z ) * i / ( e1_z - e2_z );
}

complex sec( complex z )
{
	return 1.0 / cos( z ) ;
}

complex cosec( complex z )
{
	return 1.0 / sin( z ) ;
}

//////// Hyperbolic //////////////////////////
complex cosh( complex z )
{
	return ( exp( z ) + exp( -z ) ) / 2.0 ;
}

complex sinh( complex z )
{
	return ( exp( z ) - exp( -z ) ) / 2.0 ;
}

complex tgh( complex z )
{
	complex e1_z = exp( z ) , e2_z = exp( -z ) ;
	
	return ( e1_z - e2_z ) / ( e1_z + e2_z ) ;
}

complex ctgh( complex z )
{
	complex e1_z = exp( z ) , e2_z = exp( -z ) ;

	return ( e1_z + e2_z ) / ( e1_z - e2_z );
}

complex sech( complex z )
{
	return 1.0 / cosh( z );
}

complex cosech( complex z )
{
	return 1.0 / sinh( z );
}


//////// Inverse Trigonometric //////////////////////////
complex asin( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return ln( i * z - pow( 1.0 - z * z , 0.5 ) ) / i ;
}

complex acos( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return ln( z - i * pow( 1.0 - z * z , 0.5 ) ) / i ;
}

complex atg( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return ( ln( i * z - pow( 1.0 - z * z , 0.5 ) ) ) / ( ln( z - i * pow( 1.0 - z * z , 0.5 ) ) ) ;
}

complex actg( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return ( ln( z - i * pow( 1.0 - z * z , 0.5 ) ) ) / ( ln( i * z - pow( 1.0 - z * z , 0.5 ) ) ) ;
}

complex asec( complex z )
{
	return 1.0 / acos( z ) ;
}

complex acosec( complex z )
{
	return 1.0 / asin( z ) ;
}

//////// Inverse Hyperbolic //////////////////////////
complex asinh( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return asin( i * z ) / i;
}

complex acosh( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return acos( z ) / i ;
}

complex atgh( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return asin( i * z ) / acos( z ) ;
}

complex actgh( complex z )
{
	complex i( 0.0 , 1.0 ) ;

	return acos( z ) / asin( i * z ) ;
}

complex asech( complex z )
{
	return 1.0 / acosh( z ) ;
}

complex acosech( complex z )
{
	return 1.0 / asinh( z ) ;
}

///// Miscellaneous ////////////////////////////
double complex::euclidean_dist( complex z_to )
{
    double d = ( real - z_to.real ) * ( real - z_to.real ) + ( imag - z_to.imag ) * ( imag - z_to.imag ) ;

	return sqrt( d ) ;
}

double complex::spherical_dist( complex z_to )
{
      return 2 * ( ( *this - z_to ).radius() ) / sqrt( ( 1 + radius() * radius() ) * ( 1 + z_to.radius() * z_to.radius() ) ) ;
}

complex complex::rotate( complex z, complex center, double angle )
{
      // angle value shall be set in radians
	  z -= center;
	  double r = z.radius(), a = z.angle();

	  complex rotated_point( r * cos( a + angle ), r * sin( a + angle ) ) ;

	  return ( rotated_point + center ) ;
}

char* complex::string()
{
	char* buffer = new char[52] ;

	sprintf( buffer, "(%.12f)+i*(%.12f)", real, imag ) ;
	
	return buffer ;
}

// from the Riemann sphere (diameter 1, center at the origin) to the complex plane
complex inverse_stereographic( double u, double v, double w )
{
	double out_r = u / ( 1 - w ) ;
	double out_i = v / ( 1 - w ) ;

	return complex( out_r, out_i ) ;
}

void stereographic( double x, double y, double& u, double& v, double& w )
{
	double den = x * x + y * y + 1 ;

	u = x / den , v = y / den, w = ( x * x + y * y ) / den ;
}